"""
keras==2.0.0, 2.2 not working
"""

from __future__ import division, print_function
import os
import numpy as np

import utils.model_builder
from utils.objectives import weighted_crossentropy
import utils.metrics
import keras.backend
from keras import callbacks
import keras.layers
import keras.models
import keras.optimizers

from tfutils import imread, pad_image
from patches import extract_patches, pick_coords, pick_coords_list, extract_patch_list, _extract_patches, PatchDataGeneratorList
from tfutils import make_outputdir, normalize, conv_labels2dto3d
from os.path import join
from tfutils import parse_image_files
from on_the_fly import affine_transform, random_flip, random_rotate, random_illum


FRAC_TEST = 0.1
augment_pipe = [affine_transform(points=10, distort=2), random_flip(),
                random_rotate(), random_illum(perc=0.25)]


def define_callbacks(output):
    csv_logger = callbacks.CSVLogger(join(output, 'training.log'))
    # earlystop = callbacks.EarlyStopping(monitor='loss', patience=2)
    fpath = join(output, 'weights.{epoch:02d}-{loss:.2f}-{categorical_accuracy:.2f}.hdf5')
    cp_cb = callbacks.ModelCheckpoint(filepath=fpath, monitor='loss', save_best_only=True)
    return [csv_logger, cp_cb]


def train(image_list, labels_list, output, patchsize=256, nsteps=100,
          batch_size=16, nepochs=10, weights=None, loss_weights=[1.0, 1.0, 10.0]):

    li_image, li_labels = [], []
    for image_path, labels_path in zip(image_list, labels_list):
        image, labels = imread(image_path), imread(labels_path).astype(np.uint8)
        image = normalize(image)
        if image.ndim == 2:
            image = np.expand_dims(image, -1)
        elif image.ndim == 3:
            image = np.moveaxis(image, 0, -1)
        li_image.append(image)
        li_labels.append(labels)
    num_colors = li_image[0].shape[-1]

    num_tests = int(nsteps * batch_size * FRAC_TEST)
    ecoords = pick_coords_list(nsteps * batch_size, li_labels, patchsize, patchsize)
    li_labels = [conv_labels2dto3d(lb) for lb in li_labels]
    ecoords_tests, ecoords_train = ecoords[:num_tests], ecoords[num_tests:]
    x_tests, y_tests = extract_patch_list(li_image, li_labels, ecoords_tests, patchsize, patchsize)

    model = utils.model_builder.get_model(patchsize, patchsize, num_colors, activation=None)
    loss = lambda y_true, y_pred: weighted_crossentropy(y_true, y_pred, loss_weights)
    metrics = [keras.metrics.categorical_accuracy,
               utils.metrics.channel_recall(channel=0, name="background_recall"),
               utils.metrics.channel_precision(channel=0, name="background_precision"),
               utils.metrics.channel_recall(channel=1, name="interior_recall"),
               utils.metrics.channel_precision(channel=1, name="interior_precision"),
               utils.metrics.channel_recall(channel=2, name="boundary_recall"),
               utils.metrics.channel_precision(channel=2, name="boundary_precision"),
               ]
    if weights is not None:
        model.load_weights(weights)
    optimizer = keras.optimizers.RMSprop(lr=1e-4)
    model.compile(loss=loss, metrics=metrics, optimizer=optimizer)
    # model.summary()

    make_outputdir(output)
    callbacks = define_callbacks(output)

    datagen = PatchDataGeneratorList(augment_pipe)

    history = model.fit_generator(
        generator=datagen.flow(li_image, li_labels, ecoords_train, patchsize, patchsize, batch_size=batch_size, shuffle=True),
        steps_per_epoch=nsteps,
        epochs=nepochs,
        validation_data=(x_tests, y_tests),
        # validation_steps=len(ecoords_train)/batch_size,
        validation_steps=len(ecoords_train),
        callbacks=callbacks,
        verbose=1
    )

    # score = model.evaluate(x_tests, y_tests, batch_size=batch_size)
    rec = dict(zip(model.metrics_names, [history.history[i] for i in model.metrics_names]))
    np.savez(join(output, 'records.npz'), **rec)

    json_string = model.to_json()
    open(join(output, 'cnn_model.json'), 'w').write(json_string)
    model.save_weights(join(output, 'cnn_model_weights.hdf5'))
    yaml_string = model.to_yaml()
    open(join(output, 'cnn_model.yaml'), 'w').write(yaml_string)


def _parse_command_line_args():
    """
    image:  Path to a tif or png file (e.g. data/nuc0.png).
            To pass multiple image files (size can be varied), use syntax like
            "-i im0.tif / im1.tif / im2.tif", and pass the same number of labels.
    labels: (e.g. data/labels0.tif)
    nsteps: A number of steps per epoch. A total patches passed to a model will be
            batch * nsteps.
    batch:  Typically 10-32? This will affect a memory usage.
    """

    import argparse
    parser = argparse.ArgumentParser(description='predict')
    parser.add_argument('-i', '--image', help='image file path', nargs="*")
    parser.add_argument('-l', '--labels', help='labels file path', nargs="*")
    parser.add_argument('-o', '--output', default='.', help='output directory')
    parser.add_argument('-n', '--nsteps', type=int, default=100, help='number of steps')
    parser.add_argument('-b', '--batch', type=int, default=16)
    parser.add_argument('-e', '--epoch', type=int, default=50)
    parser.add_argument('-p', '--patch', type=int, default=256,
                        help='pixel size of image patches. make it divisible by 8')
    parser.add_argument('-w', '--weights', help='hdf5 weight file path')
    parser.add_argument('-q', '--loss', type=float, nargs='+', action='append',
                        help='Background, Interior, Border weight for loss function')
    return parser.parse_args()


def _main():
    args = _parse_command_line_args()
    if args.loss is None:
        args.loss = [[1., 1., 1.]]
    images = parse_image_files(args.image)[0]
    labels = parse_image_files(args.labels)[0]
    train(images, labels, args.output, args.patch,
          args.nsteps, args.batch, args.epoch, args.weights,
          args.loss[0])


if __name__ == "__main__":
    _main()



import numpy as np
import skimage.segmentation
import skimage.io
import keras.backend as K
import tensorflow as tf

debug = False

def channel_precision(channel, name):
    def precision_func(y_true, y_pred):
        y_pred_tmp = K.cast(tf.equal( K.argmax(y_pred, axis=-1), channel), "float32")
        true_positives = K.sum(K.round(K.clip(y_true[:,:,:,channel] * y_pred_tmp, 0, 1)))
        predicted_positives = K.sum(K.round(K.clip(y_pred_tmp, 0, 1)))
        precision = true_positives / (predicted_positives + K.epsilon())

        return precision
    precision_func.__name__ = name
    return precision_func


def channel_recall(channel, name):
    def recall_func(y_true, y_pred):
        y_pred_tmp = K.cast(tf.equal( K.argmax(y_pred, axis=-1), channel), "float32")
        true_positives = K.sum(K.round(K.clip(y_true[:,:,:,channel] * y_pred_tmp, 0, 1)))
        possible_positives = K.sum(K.round(K.clip(y_true[:,:,:,channel], 0, 1)))
        recall = true_positives / (possible_positives + K.epsilon())

        return recall
    recall_func.__name__ = name
    return recall_func


## PROBMAP TO CONTOURS TO LABEL

def probmap_to_contour(probmap, threshold = 0.5):
    # assume 2D input
    outline = probmap >= threshold

    return outline

def contour_to_label(outline, image):
    # see notebook contours_to_labels for why we do what we do here

    # get connected components
    labels = skimage.morphology.label(outline, background=1)
    skimage.morphology.remove_small_objects(labels, min_size = 100, in_place = True)

    n_ccs = np.max(labels)

    # buffer label image
    filtered_labels = np.zeros_like(labels, dtype=np.uint16)

    # relabel as we don't know what connected component the background has been given before
    label_index = 1

    # start at 1 (0 is contours), end at number of connected components
    for i in range(1, n_ccs + 1):

        # get mask of connected compoenents
        mask = labels == i

        # get mean
        mean = np.mean(np.take(image.flatten(),np.nonzero(mask.flatten())))

        if(mean > 50/255):
            filtered_labels[mask] = label_index
            label_index = label_index + 1

    return filtered_labels


## PROBMAP TO PRED TO LABEL

def probmap_to_pred(probmap, boundary_boost_factor):
    # we need to boost the boundary class to make it more visible
    # this shrinks the cells a little bit but avoids undersegmentation
    pred = np.argmax(probmap * [1, 1, boundary_boost_factor], -1)

    return pred


def pred_to_label(pred, cell_min_size, cell_label=1):

    cell = (pred == cell_label)
    # fix cells
    cell = skimage.morphology.remove_small_holes(cell, min_size=cell_min_size)
    cell = skimage.morphology.remove_small_objects(cell, min_size=cell_min_size)

    # label cells only
    [label, num] = skimage.morphology.label(cell, return_num=True)
    return label


def compare_two_labels(label_model, label_gt, return_IoU_matrix):

    # get number of detected nuclei
    nb_nuclei_gt = np.max(label_gt)
    nb_nuclei_model = np.max(label_model)

    # catch the case of an empty picture in model and gt
    if nb_nuclei_gt == 0 and nb_nuclei_model == 0:
        if(return_IoU_matrix):
            return [0, 0, 1, np.empty(0)]
        else:
            return [0, 0, 1]

    # catch the case of empty picture in model
    if nb_nuclei_model == 0:
        if(return_IoU_matrix):
            return [0, nb_nuclei_gt, 0, np.empty(0)]
        else:
            return [0, nb_nuclei_gt, 0]

    # catch the case of empty picture in gt
    if nb_nuclei_gt == 0:
        if(return_IoU_matrix):
            return [nb_nuclei_model, 0, 0, np.empty(0)]
        else:
            return [nb_nuclei_model, 0, 0]

    # build IoU matrix
    IoUs = np.full((nb_nuclei_gt, nb_nuclei_model), -1, dtype = np.float32)

    # calculate IoU for each nucleus index_gt in GT and nucleus index_pred in prediction
    # TODO improve runtime of this algorithm
    for index_gt in range(1,nb_nuclei_gt+1):

        nucleus_gt = label_gt == index_gt
        number_gt = np.sum(nucleus_gt)

        for index_model in range(1,nb_nuclei_model+1):

            if debug:
                print(index_gt, "/", index_model)

            nucleus_model = label_model == index_model
            number_model = np.sum(nucleus_model)

            same_and_1 = np.sum((nucleus_gt == nucleus_model) * nucleus_gt)

            IoUs[index_gt-1,index_model-1] = same_and_1 / (number_gt + number_model - same_and_1)

    # get matches and errors
    detection_map = (IoUs > 0.5)
    nb_matches = np.sum(detection_map)

    detection_rate = IoUs * detection_map

    nb_overdetection = nb_nuclei_model - nb_matches
    nb_underdetection = nb_nuclei_gt - nb_matches

    mean_IoU = np.mean(np.sum(detection_rate, axis = 1))

    if(return_IoU_matrix):
        result = [nb_overdetection, nb_underdetection, mean_IoU, IoUs]
    else:
        result = [nb_overdetection, nb_underdetection, mean_IoU]
    return result

def splits_and_merges_3_class(y_model_pred, y_gt_pred):

    # get segmentations
    label_gt = pred_to_label(y_gt_pred, cell_min_size=2)
    label_model = pred_to_label(y_model_pred, cell_min_size=2)

    # compare labels
    result = compare_two_labels(label_model, label_gt, False)

    return result

def splits_and_merges_boundary(y_model_outline, y_gt_outline, image):

    # get segmentations
    label_gt = contour_to_label(y_gt_outline, image)
    label_model = contour_to_label(y_model_outline, image)

    # compare labels
    result = compare_two_labels(label_model, label_gt, False)

    return result



import numpy as np
import skimage.transform

# Based on example code from:
# http://scikit-image.org/docs/dev/auto_examples/transform/plot_piecewise_affine.html


def deform(image1, image2, points=10, distort=3.0):

    # create deformation grid
    rows, cols = image1.shape[0], image1.shape[1]
    src_cols = np.linspace(0, cols, points)
    src_rows = np.linspace(0, rows, points)
    src_rows, src_cols = np.meshgrid(src_rows, src_cols)
    src = np.dstack([src_cols.flat, src_rows.flat])[0]

    # add distortion to coordinates
    s = src[:, 1].shape
    dst_rows = src[:, 1] + np.random.normal(size=s)*np.random.uniform(0.0, distort, size=s)
    dst_cols = src[:, 0] + np.random.normal(size=s)*np.random.uniform(0.0, distort, size=s)

    dst = np.vstack([dst_cols, dst_rows]).T

    tform = skimage.transform.PiecewiseAffineTransform()
    tform.estimate(src, dst)

    out_rows = rows
    out_cols = cols
    out1 = skimage.transform.warp(image1, tform, output_shape=(out_rows, out_cols), mode="symmetric")
    out2 = skimage.transform.warp(image2, tform, output_shape=(out_rows, out_cols), mode="symmetric",
                                  order=0)
    return out1, out2


def resize(x, y):
    wf = 1 + np.random.uniform(-0.25, 0.25)
    hf = 1 + np.random.uniform(-0.25, 0.25)

    w,h = x.shape[0:2]

    wt, ht = int(wf*w), int(hf*h)

    new_x = skimage.transform.resize(x, (wt,ht))
    new_y = skimage.transform.resize(y, (wt,ht))

    return new_x, new_y




from __future__ import division
import numpy as np
import os
from keras.preprocessing.image import ImageDataGenerator, Iterator, array_to_img
from keras import backend as K
from utils.augmentation import deform


POINTS, DISTORT = 20, 5

def _sample_coords_weighted(num, shape, weights):
    flat_idx = np.arange(shape[0] * shape[1])
    chosen_flat = np.random.choice(flat_idx, num, p=weights/weights.sum())
    return np.unravel_index(chosen_flat, shape)


def _calc_equal_weights(features):
    ap = []
    for i in np.unique(features):
        ap.append((features == i).sum())
    frac = np.array([(np.sum(ap) - i)/np.sum(ap) for i in ap])
    prob_2d = np.zeros(features.shape)
    for i in np.unique(features):
        prob_2d[features == i] = frac[i]
    return prob_2d


def pick_coords(num, features, patch_h, patch_w):
    """
    features: img with labels
    """
    prob_2d = _calc_equal_weights(features.astype(np.uint8))
    _ph, _pw = int(np.floor(patch_h/2)), int(np.floor(patch_w/2))
    perim = np.zeros(prob_2d.shape, dtype=np.bool)
    perim[_ph:-_ph, _pw:-_pw] = True
    prob_2d[~perim] = 0
    return _sample_coords_weighted(num, features.shape, prob_2d.flatten())


def pick_coords_list(num, li_features, patch_h, patch_w):
    """
    features: img with labels
    """
    from random import shuffle
    num = int(num/len(li_features))
    li_coords = []
    for en, features in enumerate(li_features):
        coords = pick_coords(num, features, patch_h, patch_w)
        coords = zip(*(np.ones(num, np.uint8) * en, coords[0], coords[1]))
        li_coords.extend(coords)
    shuffle(li_coords)
    return li_coords



# def _pick_coords(num, features, patch_h, patch_w):
#     """
#     features: img with labels
#     """
#     prob_2d = _calc_equal_weights(features.astype(np.uint8))
#     _ph, _pw = int(np.floor(patch_h/2)), int(np.floor(patch_w/2))
#     perim = np.zeros(prob_2d.shape, dtype=np.bool)
#     perim[_ph:-_ph, _pw:-_pw] = True
#     prob_2d[~perim] = 0
#     return _sample_coords_weighted(num, features.shape, prob_2d.flatten())


def extract_patches(num, x, y, patch_h, patch_w):
    """
    x: input images
    y: feature image with labels

    Sample many windows from a large image. It will correct for labels unbalance.
    (If there are less labels for cell boundaries, it increases the sampling probability)
    """
    coords = pick_coords(num, y, patch_h, patch_w)
    h, w = int(np.floor(patch_h/2)), int(np.floor(patch_w/2))
    xstack = np.zeros((num, patch_h, patch_w, x.shape[-1]), np.float32)
    ystack = np.zeros(num)
    for n, (ch, cw) in enumerate(zip(*coords)):
        xstack[n, :, :] = x[ch-h:ch+h+1, cw-w:cw+w+1]
        ystack[n] = y[ch, cw]
    return xstack, ystack


def _extract_patches(x, y, coords, patch_h, patch_w):
    """
    x: input images x.shape = (hpix, wpix, ch)
    y: feature image with labels

    Sample many windows from a large image. It will correct for labels unbalance.
    (If there are less labels for cell boundaries, it increases the sampling probability)
    """
    h, w = int(np.floor(patch_h/2)), int(np.floor(patch_w/2))
    xstack = np.zeros((len(coords), patch_h, patch_w, x.shape[-1]), np.float32)
    ystack = np.zeros((len(coords), patch_h, patch_w, y.shape[-1]), np.float32)
    for n, (ch, cw) in enumerate(coords):
        xstack[n, :, :, :] = x[ch-h:ch+h, cw-w:cw+w]
        ystack[n, :, :, :] = y[ch-h:ch+h, cw-w:cw+w]
    return xstack, ystack


def extract_patch_list(lix, liy, ecoords, patch_h, patch_w):
    h, w = int(np.floor(patch_h/2)), int(np.floor(patch_w/2))
    xstack = np.zeros((len(ecoords), patch_h, patch_w, lix[0].shape[-1]), np.float32)
    ystack = np.zeros((len(ecoords), patch_h, patch_w, liy[0].shape[-1]), np.float32)
    for n, (cn, ch, cw) in enumerate(ecoords):
        xstack[n, :, :, :] = lix[cn][ch-h:ch+h, cw-w:cw+w]
        ystack[n, :, :, :] = liy[cn][ch-h:ch+h, cw-w:cw+w]
    return xstack, ystack


class PatchDataGenerator(ImageDataGenerator):
    def flow(self, x, y, coords, patch_h, patch_w, batch_size=32, shuffle=True, seed=None,
             save_to_dir=None, save_prefix='', save_format='png'):
        return CropIterator(
            x, y, self, coords=coords, patch_h=patch_h, patch_w=patch_w,
            batch_size=batch_size, shuffle=shuffle, seed=seed,
            data_format=self.data_format,
            save_to_dir=save_to_dir, save_prefix=save_prefix, save_format=save_format)


class CropIterator(Iterator):
    def __init__(self, x, y, image_data_generator, coords, patch_h, patch_w,
                 batch_size=32, shuffle=False, seed=None, func_patch=_extract_patches,
                 data_format=None, save_to_dir=None, save_prefix='',
                 save_format='png', aug_pipeline=()):
        self.coords = coords
        self.patch_h = patch_h
        self.patch_w = patch_w
        self.aug_pipeline = aug_pipeline
        if isinstance(x, list):
            self._x, self._y = x[:], y[:]
            chnum = self._x[0].shape[-1]
        else:
            self._x, self._y = x.copy(), y.copy()
            self._x = self._x[0, :, :, :]
            chnum = self._x.shape[-1]
        self.x = np.asarray(np.zeros((1, patch_h, patch_w, chnum)), dtype=K.floatx())
        if y is not None:
            self.y = np.asarray(np.zeros((1, patch_h, patch_w, self._y[0].shape[-1])), dtype=K.floatx())
        else:
            self.y = None
        self.n = len(self.coords)
        self.func_patch = func_patch

        self.image_data_generator = image_data_generator
        self.data_format = data_format
        self.save_to_dir = save_to_dir
        self.save_prefix = save_prefix
        self.save_format = save_format
        super(CropIterator, self).__init__(len(self.coords), batch_size, shuffle, seed)

    def _get_batches_of_transformed_samples(self, index_array):
        batch_x = np.zeros(tuple([len(index_array)] + list(self.x.shape)[1:]), dtype=K.floatx())
        batch_y = np.zeros(tuple([len(index_array)] + list(self.y.shape)[1:]), dtype=K.floatx())

        batch_coords = [self.coords[i] for i in index_array]
        x, y = self.func_patch(self._x, self._y, batch_coords, self.patch_h, self.patch_w)
        # y = np.expand_dims(y, -1)
        self.x = x
        self.y = y
        index_array = np.arange(len(self.y))

        for i, j in enumerate(index_array):
            x = self.x[j]
            y = self.y[j]

            for aug in self.aug_pipeline:
                x, y = aug(x, y)
            batch_x[i] = x
            batch_y[i] = y

        if self.save_to_dir:
            for i, j in enumerate(index_array):
                img = array_to_img(batch_x[i], self.data_format, scale=True)
                fname = '{prefix}_{index}_{hash}.{format}'.format(prefix=self.save_prefix, index=j,
                hash=np.random.randint(1e4), format=self.save_format)
            img.save(os.path.join(self.save_to_dir, fname))
        return batch_x, batch_y

    def next(self):
        with self.lock:
            index_array = next(self.index_generator)
        if isinstance(index_array, tuple):
            index_array = index_array[0]
        return self._get_batches_of_transformed_samples(index_array)


class PatchDataGeneratorList(ImageDataGenerator):
    def flow(self, x, y, coords, patch_h, patch_w, aug_pipeline=(), batch_size=32, shuffle=True, seed=None,
             save_to_dir=None, save_prefix='', save_format='png'):
        return CropIterator(
            x, y, self, coords=coords, patch_h=patch_h, patch_w=patch_w,
            batch_size=batch_size, shuffle=shuffle, seed=seed,
            data_format=self.data_format, func_patch=extract_patch_list,
            save_to_dir=save_to_dir, save_prefix=save_prefix,
            save_format=save_format, aug_pipeline=aug_pipeline)




import os
import imp
import numpy as np
from scipy.ndimage import imread as imread0
import tifffile as tiff



def conv_labels2dto3d(labels):
    lbnums = np.unique(labels)
    arr = np.zeros((labels.shape[0], labels.shape[1], len(lbnums)), np.uint8)
    for i in lbnums:
        arr[:, :, i] = labels == i
    return arr


def normalize(orig_img):
    percentile = 99.9
    high = np.percentile(orig_img, percentile)
    low = np.percentile(orig_img, 100-percentile)
    img = np.minimum(high, orig_img)
    img = np.maximum(low, img)
    img = (img - low) / (high - low)
    return img


def make_outputdir(output):
    try:
        os.makedirs(output)
    except:
        pass


def imread_check_tiff(path):
    img = imread0(path)
    if img.dtype == 'object' or path.endswith('tif'):
        img = tiff.imread(path)
    return img


def imread(path):
    if isinstance(path, tuple) or isinstance(path, list):
        st = []
        for p in path:
            st.append(imread_check_tiff(p))
        img = np.dstack(st)
        if img.shape[2] == 1:
            np.squeeze(img, axis=2)
        return img
    else:
        return imread_check_tiff(path)


def parse_image_files(inputs):
    if "/" not in inputs:
        return (inputs, )
    store = []
    li = []
    while inputs:
        element = inputs.pop(0)
        if element == "/":
            store.append(li)
            li = []
        else:
            li.append(element)
    store.append(li)
    return zip(*store)


def pad_image(image):
    # assumes the image is an np array of dimensions d0 x d1 x d2 x d3
    # where d1 is height, d2 is width, and d3 is colors
    # returns a list of padded image, hpadding (tuple), wpadding (tuple)

    height = image.shape[1]
    hdelta = 8 - (height % 8)
    if (hdelta == 8):
        hpadding = (0, 0)
    elif (hdelta % 2) == 0:
        hpadding = (int(hdelta/2.0), int(hdelta/2.0))
    else:
        hpadding = (int(hdelta/2.0), int(hdelta/2.0)+1)

    width = image.shape[2]
    wdelta = 8 - (width % 8)
    if (wdelta == 8):
        wpadding = (0, 0)
    elif (wdelta % 2) == 0:
        wpadding = (int(wdelta/2.0), int(wdelta/2.0))
    else:
        wpadding = (int(wdelta/2.0), int(wdelta/2.0)+1)

    return [np.pad(image, ((0, 0), hpadding, wpadding, (0, 0)), 'constant', constant_values=0.0), hpadding, wpadding]


def normalize_predictions(predictions):
    # predictions is typically a 3 x h x w list
    predictions = np.array(predictions)
    num_rows = predictions.shape[1]
    num_cols = predictions.shape[2]

    for i in range(num_rows):
        for j in range(num_cols):
            prediction = predictions[:, i, j]
            if (prediction.sum() != 0):
                predictions[:, i, j] = prediction / prediction.sum()

    return predictions

            




import keras.metrics
import tensorflow as tf


def weighted_crossentropy(y_true, y_pred, weights):

    class_weights = tf.constant([[[[weights[0], weights[1], weights[2]]]]])

    unweighted_losses = tf.nn.softmax_cross_entropy_with_logits_v2(labels=y_true, logits=y_pred)

    weights = tf.reduce_sum(class_weights * y_true, axis=-1)

    weighted_losses = weights * unweighted_losses

    loss = tf.reduce_mean(weighted_losses)

    return loss




from __future__ import division
import os
import numpy as np
from os.path import join, basename, splitext
from tfutils import imread
from tfutils import make_outputdir, normalize
from tfutils import pad_image, normalize_predictions
import tifffile as tiff
import utils.model_builder


def predict(img_path, weight_path):
    x = imread(img_path)
    x = normalize(x)

    if x.ndim == 2:
        x = np.expand_dims(x, -1)
    elif x.ndim == 3:
        x = np.moveaxis(x, 0, -1)
    x = np.expand_dims(x, 0)
    num_colors = x.shape[-1]

    x, hpadding, wpadding = pad_image(x)

    model = utils.model_builder.get_model(x.shape[1], x.shape[2], num_colors, activation=None)
    model.load_weights(weight_path)
    predictions = model.predict(x, batch_size=1)
    predictions = [predictions[0, :, :, i] for i in range(predictions.shape[-1])]

    # resize predictions to match image dimensions (i.e. remove padding)
    height = predictions[0].shape[0]
    width = predictions[0].shape[0]
    predictions = [p[hpadding[0]:height-hpadding[1], wpadding[0]:width-wpadding[1]] for p in predictions]

    predictions = normalize_predictions(predictions)

    return predictions


def save_output(outputdir, images, pattern):
    make_outputdir(outputdir)
    for num, img in enumerate(images):
        tiff.imsave(join(outputdir, '{0}_l{1}.tif'.format(pattern, num)), img)


def _parse_command_line_args():
    import argparse
    parser = argparse.ArgumentParser(description='predict')
    parser.add_argument('-i', '--image', help='image file path')
    parser.add_argument('-w', '--weight', help='hdf5 file path')
    parser.add_argument('-o', '--output', default='.', help='output directory')
    return parser.parse_args()


def _main():
    args = _parse_command_line_args()
    images = predict(args.image, args.weight)
    save_output(args.output, images, splitext(basename(args.image))[0])


if __name__ == "__main__":
    _main()



from utils.augmentation import deform
import numpy as np


def affine_transform(points=20, distort=3):
    def func(x, y):
        return deform(x, y, points, distort)
    return func


def random_flip():
    def func(x, y):
        rand_flip = np.random.randint(low=0, high=2)
        if(rand_flip == 0):
            x = np.flip(x, 0)
            y = np.flip(y, 0)
        return x, y
    return func


def random_rotate():
    def func(x, y):
        rand_rotate = np.random.randint(low=0, high=4)
        for _ in range(rand_rotate):
            x = np.rot90(x)
            y = np.rot90(y)
        return x, y
    return func


def random_illum(perc=0.25):
    def func(x, y):
        ifactor = 1 + np.random.uniform(perc-1, 1-perc)
        x *= ifactor
        return x, y
    return func



import numpy as np
from tfutils import parse_image_files
from tfutils import normalize, conv_labels2dto3d, make_outputdir
from patches import extract_patches, pick_coords, pick_coords_list, extract_patch_list, _extract_patches, PatchDataGeneratorList
from tfutils import imread
from on_the_fly import affine_transform, random_flip, random_rotate, random_illum
import tifffile as tiff
from os.path import join


def train(image_list, labels_list, output, patchsize=256, nsamples=10):
    li_image, li_labels = [], []
    for image_path, labels_path in zip(image_list, labels_list):
        image, labels = imread(image_path), imread(labels_path).astype(np.uint8)
        image = normalize(image)
        if image.ndim == 2:
            image = np.expand_dims(image, -1)
        elif image.ndim == 3:
            image = np.moveaxis(image, 0, -1)
        li_image.append(image)
        li_labels.append(labels)

    ecoords = pick_coords_list(nsamples, li_labels, patchsize, patchsize)
    li_labels = [conv_labels2dto3d(lb) for lb in li_labels]
    x_tests, y_tests = extract_patch_list(li_image, li_labels, ecoords, patchsize, patchsize)

    from train import augment_pipe
    make_outputdir(output)

    for i in range(nsamples):
        x, y = x_tests[i], y_tests[i]
        for aug in augment_pipe:
            x, y = aug(x, y)
        tiff.imsave(join(output, 'x{0:04}.tif'.format(i)), x.astype(np.float32))
        tiff.imsave(join(output, 'y{0:04}.tif'.format(i)), y.astype(np.uint16))




def _parse_command_line_args():
    """
    image:  Path to a tif or png file (e.g. data/nuc0.png).
            To pass multiple image files (size can be varied), use syntax like
            "-i im0.tif / im1.tif / im2.tif", and pass the same number of labels.
    labels: (e.g. data/labels0.tif)
    n:      A number of pixels for training. Use a large number (like 1,000,000)
    batch:  Typically 128-512?
    """

    import argparse
    parser = argparse.ArgumentParser(description='predict')
    parser.add_argument('-i', '--image', help='image file path', nargs="*")
    parser.add_argument('-l', '--labels', help='labels file path', nargs="*")
    parser.add_argument('-o', '--output', default='.', help='output directory')
    parser.add_argument('-n', '--nsamples', type=int, default=10, help='number of samples')
    parser.add_argument('-p', '--patch', type=int, default=256,
                        help='pixel size of image patches. make it odd')
    return parser.parse_args()


def _main():
    args = _parse_command_line_args()
    images = parse_image_files(args.image)[0]
    labels = parse_image_files(args.labels)[0]
    train(images, labels, args.output, args.patch, args.nsamples)

if __name__ == "__main__":
    _main()



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['cellunet', 'cellunet.utils']

package_data = \
{'': ['*'], 'cellunet': ['data/*', 'etc/*']}

install_requires = \
['keras==2.0.0',
 'numpy==1.14.6',
 'pandas==0.23.0',
 'scipy==1.1.0',
 'tensorflow==1.8.0',
 'tifffile>=2019.1,<2020.0']

setup_kwargs = {
    'name': 'cellunet',
    'version': '0.1.1',
    'description': '',
    'long_description': None,
    'author': 'Takamasa Kudo',
    'author_email': 'kudo@stanford.edu',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.5,<4.0',
}


setup(**setup_kwargs)



import keras.layers
import keras.models
import tensorflow as tf

CONST_DO_RATE = 0.5

option_dict_conv = {"activation": "relu", "border_mode": "same"}
option_dict_bn = {"mode": 0, "momentum" : 0.9}


# returns a core model from input to 64 channels of the same size
def get_core(dim1, dim2, dim3):

    # assume dim1 x dim2 image with dim3 color channels

    x = keras.layers.Input(shape=(dim1, dim2, dim3))

    a = keras.layers.Convolution2D(64, 3, 3, **option_dict_conv)(x)
    a = keras.layers.BatchNormalization(**option_dict_bn)(a)

    a = keras.layers.Convolution2D(64, 3, 3, **option_dict_conv)(a)
    a = keras.layers.BatchNormalization(**option_dict_bn)(a)


    y = keras.layers.MaxPooling2D()(a)

    b = keras.layers.Convolution2D(128, 3, 3, **option_dict_conv)(y)
    b = keras.layers.BatchNormalization(**option_dict_bn)(b)

    b = keras.layers.Convolution2D(128, 3, 3, **option_dict_conv)(b)
    b = keras.layers.BatchNormalization(**option_dict_bn)(b)


    y = keras.layers.MaxPooling2D()(b)

    c = keras.layers.Convolution2D(256, 3, 3, **option_dict_conv)(y)
    c = keras.layers.BatchNormalization(**option_dict_bn)(c)

    c = keras.layers.Convolution2D(256, 3, 3, **option_dict_conv)(c)
    c = keras.layers.BatchNormalization(**option_dict_bn)(c)


    y = keras.layers.MaxPooling2D()(c)

    d = keras.layers.Convolution2D(512, 3, 3, **option_dict_conv)(y)
    d = keras.layers.BatchNormalization(**option_dict_bn)(d)

    d = keras.layers.Convolution2D(512, 3, 3, **option_dict_conv)(d)
    d = keras.layers.BatchNormalization(**option_dict_bn)(d)


    # UP

    d = keras.layers.UpSampling2D()(d)

    y = keras.layers.merge([d, c], concat_axis=3, mode="concat")

    e = keras.layers.Convolution2D(256, 3, 3, **option_dict_conv)(y)
    e = keras.layers.BatchNormalization(**option_dict_bn)(e)

    e = keras.layers.Convolution2D(256, 3, 3, **option_dict_conv)(e)
    e = keras.layers.BatchNormalization(**option_dict_bn)(e)

    e = keras.layers.UpSampling2D()(e)


    y = keras.layers.merge([e, b], concat_axis=3, mode="concat")

    f = keras.layers.Convolution2D(128, 3, 3, **option_dict_conv)(y)
    f = keras.layers.BatchNormalization(**option_dict_bn)(f)

    f = keras.layers.Convolution2D(128, 3, 3, **option_dict_conv)(f)
    f = keras.layers.BatchNormalization(**option_dict_bn)(f)

    f = keras.layers.UpSampling2D()(f)


    y = keras.layers.merge([f, a], concat_axis=3, mode="concat")

    y = keras.layers.Convolution2D(64, 3, 3, **option_dict_conv)(y)
    y = keras.layers.BatchNormalization(**option_dict_bn)(y)

    y = keras.layers.Convolution2D(64, 3, 3, **option_dict_conv)(y)
    y = keras.layers.BatchNormalization(**option_dict_bn)(y)

    return [x, y]

def get_model(dim1, dim2, dim3, activation="softmax"):

    [x, y] = get_core(dim1, dim2, dim3)

    y = keras.layers.Convolution2D(3, 1, 1, **option_dict_conv)(y)

    if activation is not None:
        y = keras.layers.Activation(activation)(y)

    model = keras.models.Model(x, y)

    return model